from cli.commandlineparser import CommandLineParser
from cli.defaultconfig import DefaultConfig
from control.controlstruct import ControlStruct
from control.jandal import JandalError

from parser import *

def main():
    defaultConfig = DefaultConfig()
    cliparser = CommandLineParser()
    if cliparser.showConfig():
        cliparser.downloadStandardConfig()

    if cliparser.showInfo():
        cliparser.displayInfo()

    # cliparser.installCssEngine()
    cssEngine = cliparser.getCssEngine()

    confnames = []
    innames = cliparser.getInputFiles()
    outdirname = cliparser.getOutputDir()
    if cliparser.getUserConfig():
        confnames.append(cliparser.getUserConfig())

    if not innames and not cliparser.showInfo() and not cliparser.showConfig():
        raise RuntimeError("Error: no input files provided...")

    conf = parseconf(confnames, defaultConfig)  # this function parses the configuration filenames in the list and returns config files

    # TODO: verify this, might not need this
    # if outdirname is not None: 
    #   if not os.path.isdir(outdirname) and len(innames) > 1:
    #     raise RuntimeError('cannot handle one outfile with multiple infiles')

    # for each input file, format the corresponding output file
    for inname in innames:
        filename = re.sub(r'.jemdoc$', '', inname) + '.html'
        if outdirname is None:
            outdirname = '.'
            outfile = filename
        elif os.path.isdir(outdirname):
            # if directory, prepend directory to automatically generated name.
            outfile = os.path.join(outdirname, filename)
        # else:
        #     outfile = outdirname

        infile = open(inname, 'rUb')
        outfile = open(outfile, 'w')

        # create a control struct with the necessary parsed configuration
        print "Processing \033[96m%s...\033[00m" % inname
        f = ControlStruct(infile, outfile, conf, inname, outdirname)
        procfile(f, cliparser)


# main
if __name__ == '__main__':
    main()



#!/usr/bin/env python2
import os
import subprocess
import sys
import shutil
from bs4 import BeautifulSoup


class ReactFactory:
    def __init__(self):
        name, title = self.getMetaDataFromUser()
        self._appTitle = title
        self._reactDirname = name
        self._components = []

    def getMetaDataFromUser(self):
        print """
\033[96mJemdoc 2.0 Initialization\033[00m

This utility will walk you through creating a react application.
If no answer is provided by the user a default name in parenthesis will be used
Press ^C \033[91m(Ctrl + C)\033[00m at any time to quit.
"""

        prompt = "application name (\033[96mjemdoc-website\033[00m): "
        defaultName = "jemdoc-website"
        name = raw_input(prompt)
        if name == "" or not name:
            name = defaultName.strip().lower()

        prompt = "application title (\033[96mJemdoc Website\033[00m): "
        defaultTitle = "Jemdoc Website"
        title = raw_input(prompt)
        if title == "" or not title:
            title = defaultTitle.strip().lower()
        
        return (name.strip().lower(), title.strip().capitalize())

    def requestToCreateReactApp(self):
        response = raw_input(
            "Would you like to install create-react-app (\033[92my\033[00m/\033[91mn\033[00m) ").lower()
        while response not in ('y', 'yes', 'yeah', 'yah', 'n', 'no', 'nah'):
            response = raw_input(
                "Would you like to install create-react-app (\033[92my\033[00m/\033[91mn\033[00m) ").lower()

        if response in ('yes', 'y', 'yah', 'yeah'):
            try:
                output = subprocess.check_output(
                    'npm i create-react-app', shell=True)
                return True
            except subprocess.CalledProcessError:
                print "The CLI `create-react-app unavailable`\n Creating a non-react jemdoc website..."
                return False
        else:
            print "The CLI `create-react-app` is needed to create a react application.\nCreating a non-react jemdoc website..."
            return False

    def createReactApp(self):
        output = subprocess.check_output(
            "create-react-app %s" % self._reactDirname, shell=True)

    def testCreateReactApp(self):
        print "Creating react application named '\033[96m%s\033[00m'" % self._reactDirname
        try:
            self.createReactApp()
            return True
        except subprocess.CalledProcessError as ex:
            print "\033[91mWarn\033[00m: Unable to create react application without create-react-app"
            return self.requestToCreateReactApp()

    def testNPM(self):
        try:
            output = subprocess.check_output('npm -v', shell=True)
            return True
        except subprocess.CalledProcessError as ex:
            print "\033[91mWarn\033[00m: NPM needs to be installed to compile as a React Application\nCreating a non-react jemdoc website..."
            return False

    def install(self, package):
        try:
            print "Installing %s" %package
            subprocess.check_output(
                [sys.executable, "-m", "pip2", "-q", "install", "--user", package])
            return True
        except subprocess.CalledProcessError as ex:
            print "\033[91mWarn\033[00m: %s cannot be installed.\n Creating a non-react jemdoc website..." %package
            return False

    def testBS4(self):
        try:
            from bs4 import BeautifulSoup
            return True
        except ImportError:
            installed = self.install("bs4")
            return installed

    def isDuplicateDirname(self):
        """
        This method checks if a duplicate name exists in the directory in which the react application is to be created.
        If duplicate name exists, returns True. Otherwise returns False

        Input: void
        Output: boolean
        """
        dirs = os.listdir(os.getcwd())
        if self._reactDirname in dirs:  # if has duplicate name
            return True
        else:
            return False

    def overwriteDuplicateFolderName(self):
        """
        This method checks if user wants to overwrite duplicate directory name if it exists. User is asked and based on user response, erases duplicate 
        """
        print "\033[92mInfo\033[00m: Directory named '%s' already exists." % self._reactDirname
        prompt = "Would you like to overwrite directory (\033[92my\033[00m/\033[91mn\033[00m)? "
        response = raw_input(prompt).lower()
        while response not in ('y', 'yes', 'yeah', 'yah', 'n', 'no', 'nah'):
            response = raw_input(prompt).lower()

        if response in ('yes', 'y', 'yah', 'yeah'):
            print "\033[92mInfo\033[00m: Deleting directory named '%s'" %self._reactDirname
            path = os.path.join(os.getcwd(), self._reactDirname)
            shutil.rmtree(path)
            return True
        else:
            return False

    def testRequirements(self):
        """
        This method verifies that all the necessary requirements needed to create a react application have been met
        If True a react application is created, otherwise it falls back to creating a static jemdoc website

        Input: void
        Output: boolean - whether all the requirements have been met
        """
        if self.isDuplicateDirname():
            if not self.overwriteDuplicateFolderName():
                sys.exit(0)

        npm = self.testNPM()
        reactApp = self.testCreateReactApp() # this method also creates the react application if create-react-app is available
        soup4 = self.testBS4()

        if (npm and reactApp and soup4):
            print "Creating react application \033[91m(Ctrl + c to interrupt)\033[00m..."
            return True
        else:
            print "System requirements have not been met. Please ensure NPM and create-react-app are installed on system"
            print "Creating a non-react jemdoc website..."
            return False

    def raiseError(self, msg):
        print msg
        sys.exit(0)

    def getComponentsDir(self):
        return os.path.join(os.getcwd(), self._reactDirname, "src", "components")

    def getCSSDir(self):
        return os.path.join(self.getSrcDir(), "styles")

    def getSrcDir(self):
        return os.path.join(os.getcwd(), self._reactDirname, "src")

    def createComponentDir(self):
        dirname = self.getComponentsDir()
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        # os.chdir(dirname)
        # print os.getcwd()

    def createCSSDir(self):
        dirname = self.getCSSDir()
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def move(self, source, destination):
        # print destination
        try:
            shutil.move(source, destination)
            return
        except:
            os.rename(source, destination)

    def moveToCssDir(self, cssFiles):
        """
        This method moves css fileObjs provided by the user into the styles directory in the react application

        Input: string[] - names of the css fileObjs
        Return: void
        """
        # cssFiles is a list of css fileObjs
        cssDir = self.getCSSDir()
        for css in cssFiles:
            source = os.path.join(os.getcwd(), css)
            dest = os.path.join(cssDir, css)
            # print "Moving %s to %s" %(source, dest) # debug
            self.move(source, dest)

    def askForNewFilename(self):
        response = raw_input("Please enter a new fileObjname: ").capitalize()
        return response

    def addComponent(self, component):
        """
        This method adds the component fileObj to the appropriate directory
        """
        path = self.getComponentsDir()
        fileObjpath = os.join(path, component)
        while os.path.exists(fileObjpath):
            print "Component with name '%s' already exists"
            response = self.askForNewFilename()
            fileObjpath = os.join(path, response)
        os.makedirs(fileObjpath)

    def getAllCSSImports(self):
        """
        This method creates all of the css import statements for the react application
        """
        imports = ''
        cssDir = self.getCSSDir()
        # print "CSS Directory: %s" % cssDir
        cssFiles = os.listdir(cssDir)
        for css in cssFiles:
            cssPath = os.path.join(".", "styles", css)
            statement = "import '%s';\n" % cssPath
            imports += statement
        # print imports  # for debugging
        return imports

    def createComponent(self, componentName, render, state="", write=False):
        """
        This method creates a generic component and writes the component to a JSX fileObj in the  components directory
        """
        component = """
%s

import { React } from 'react';

export class %s extends React.Component {
    constructor(props) {
        super(props);

        this.state = {%s};
    }

    render() {
        return (
            <React.Fragment>
                %s
            </React.Fragment>
        );
    }
}
""" % (self.getAllCSSImports(), componentName, state, render)

        fileObjname = "%s.jsx" %componentName.capitalize()
        # print component #debug
        if write:
            self.writeComponent(fileObjname, component)
        self._components.append(componentName.capitalize())
        return component

    def writeComponent(self, fileObjname, component):
        """
        This method creates and flushes the component fileObj into the components directory as a JSX fileObj

        Input: string - fileObjname, string - component JSX fileObj contents
        Output: void
        """
        componentsDir = self.getComponentsDir()
        if not os.path.exists(componentsDir):
            os.makedirs(componentsDir)

        fileObjPath = os.path.join(componentsDir, fileObjname)
        fileObj = open(fileObjPath, "w+")
        fileObj.write(component)
        fileObj.flush()
        fileObj.close()

    def addBootstrap(self, version = "4.3.1"):
        """
        This method adds bootstrap@4.3.1, jquery@1.9.1 and popper.js@1.14.7 to the react application via npm. 
        If npm is not installed or bootstrap is unavailable the react application cannot use bootstrap

        Input: string - version number of bootstrap. Defaults to version 4.3.1
        Output: void
        """
        originalDir = os.getcwd()
        reactAppDir = os.path.join(originalDir, self._reactDirname)
        os.chdir(reactAppDir);
        try:
            print "Adding Twitter Bootstrap4 to application..."
            output = subprocess.check_output("npm i -S tsutils@3.17.1 jquery@1.9.1 popper.js@1.14.7 bootstrap@%s" %version, shell=True)
            # output = subprocess.check_output("npm i -S jquery@1.9.1", shell=True)
            # output = subprocess.check_output("npm i -S popper.js@1.14.7", shell=True)
            return True
        except subprocess.CalledProcessError as ex:
            print "\033[91mWarn\033[00m: Twitter Bootstrap4 has not been installed. Proceeding without Twitter Bootstrap4"
            return False
        finally:
            os.chdir(originalDir)

    def getBootstrapImport(self):
        """
        This method returns all the necessary reactjs Twitter Bootstrap import statements

        Input: void
        Output: string - import statements
        """
        return """
import '../node_modules/bootstrap/dist/css/bootstrap.min.css';
import '../node_modules/bootstrap/dist/js/bootstrap.min.js';
import '../node_modules/jquery/dist/jquery.slim.min.js';

"""

    # def insertBootstrap4(self):
    #     if self.addBootstrap():
    #         indexFile = os.path.join(self.getSrcDir(), "index.js")
    #         # prepend at beginning of fileObj
    #         with open(indexFile, "r+") as f:
    #             content = f.read()
    #             f.seek(0, 0)
    #             f.write(self.getBootstrapImport() + content)

    def cleanReactAppDir(self):
        """
        This method clears the predefined css fileObjs and the index.js fileObj. 
        It replaces each of these fileObjs with user created/defined fileObjs

        Input: void
        Output: void
        """
        
        # deleting the default CSS and JS fileObjs
        path = self.getSrcDir()
        fileObjs = os.listdir(path)

        for fileObj in fileObjs:
            extension = os.path.splitext(fileObj)[1]
            # print extension # debug

            if extension in (".css", ".js", ".svg"):
                fileObjpath = os.path.join(path, fileObj)
                os.remove(fileObjpath)
        

    # def insertCSSToFile(self, fileObjname, css):
    #     """
    #     This method inserts the css fileObjs into the styles directory that will be used by all components of the react application.
    #     These will be imported into all the component classes via another method
    #     """
    #     pass

    def getPublicDir(self):
        return os.path.join(os.getcwd(), self._reactDirname, "public")

    def processHTMLEntryPoint(self):
        """
        This method inserts the MathJax CDN into index.html in the react application
        Uses BeautifulSoup to insert script tags into the index.html fileObj

        Input: void
        Output: void
        """
        # creating mathjax script and config tags to insert into index.html
        path = os.path.join(self.getPublicDir(), "index.html")
        html = open(path, "r+")
        soup = BeautifulSoup(html, "html.parser")
        scriptTag = soup.new_tag('script', src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/latest.js?config=TeX-MML-AM_CHTML", type="text/javascript", asyn="async")
        configTag = soup.new_tag("script", type="text/x-mathjax-config")
        config = """
MathJax.Hub.Config({tex2jax: {
    inlineMath: [ ["$","$"], ["\\(","\\)"] ],
    displayMath: [ ["$$","$$"], ["\\[","\\]"] ]
}});"""
        configTag.append(config)
        soup.head.append(configTag)
        soup.head.append(scriptTag)


        # changing app title
        soup.title.string.replace_with(self._appTitle)

        html.close()
        
        # writing new html back to fileObj
        newHTML = soup.prettify("utf-8")
        with open(path, "wb") as fileObj:
            fileObj.write(newHTML)
        
    
    def getAllComponentImports(self):
        imports = ""
        for component in self._components:
            imports += "import { %s } from './components/%s.jsx'\n" %(component, component)


    # TODO: implement this method
    def renderToDOM(self, components):
        """
        This method renders the components provided in the list 'components' to the DOM in the order in which the list is provided

        Input: string[] - formatted component string to render. contents are rendered as is given
        Output: void
        """

        # TODO: import all components here
        # TODO: import all CSS here 
        ui = """
import React from 'react';
import ReactDOM from 'react-dom';
import './index.css';
%s

ReactDOM.render(
    <React.Fragment>
        %s
    </React.Fragment>, document.getElementById('root')
);
""" %(self.getAllComponentImports(), components)

        path = os.path.join(self.getSrcDir(), "index.js")
        fileObj = open(path, "w")
        bootstrap4 = self.getBootstrapImport()
        fileObj.write(bootstrap4)
        fileObj.write(ui)
        fileObj.flush()
        fileObj.close()



# TODO: consider when user uses multiple html pages

def seedFiles(cssFiles):
    if os.path.exists("jemdoc-website"):
        path = os.path.join(os.getcwd(), "jemdoc-website")
        try:
            shutil.rmtree(path)
        except:
            os.rmdir(path)            

    for css in cssFiles:
        if not os.path.exists(css):
            fileObj = open(css, "w")
            fileObj.flush()
            fileObj.close()

if __name__ == "__main__":
    # try-catch block is to handle the keyboard interrupt exception
    try:
        # commented means it works correctly and has been unit tested
        cssFiles = ["main.css", "textbox.css"]
        seedFiles(cssFiles)

        factory = ReactFactory()
        factory.testRequirements()
        factory.createComponentDir()
        factory.createCSSDir()
        factory.moveToCssDir(cssFiles)
        factory.cleanReactAppDir()
        factory.createComponent("TestComponent", "<h1>David</h1>", "", True) # error here
        factory.processHTMLEntryPoint()
        factory.renderToDOM("""
# <TestComponent>
#     <NonExistingComponent>Hello, World!</NonExistingComponent>
# </TestComponent>
# """)
    except KeyboardInterrupt:
        print "\n\nCancelling creation of react application...\n"
        if os.path.exists("jemdoc-website"):
            path = os.path.join(os.getcwd(), "jemdoc-website")
            try:
                shutil.rmtree(path)
            except:
                os.rmdir(path) 
    except AttributeError as ex:
        print "Attribute Error...\n"
        print ex


#TODO: fallback to using regular jemdoc in the event of an error at compile time



#!/usr/bin/env python2

import sys
import os
import re
import time
import StringIO
import subprocess # was from subprocess import *
import tempfile

from cli.commandlineparser import CommandLineParser
from control.controlstruct import ControlStruct
from control.jandal import JandalError
from styling.style import Style

def raisejandal(msg, line=0):
    if line == 0:
        s = "%s" % msg
    else:
        s = "line %d: %s" % (line, msg)
    raise JandalError(s)


def readnoncomment(f):
    l = f.readline()
    if l == '':
        return l
    elif l[0] == '#':
        return readnoncomment(f)  # read next line if current line is a comment
    else:
        return l.rstrip() + '\n'  # leave just one \n and no spaces etc.


def parseconf(cns, defaultConfig):
    syntax = {}
    warn = False
    # manually add the defaults as a file handle.
    standardconf = defaultConfig.getStandardConfig()
    fs = [ StringIO.StringIO(standardconf) ]  # fs represents all the configuration files, including the default configuration file, as file objects
    for sname in cns:
        fs.append(open(sname, 'rb'))  # opens each configuration file and appends the file object to fs

    for f in fs:  # for each file object in fs
        # ControlStruct(fileObj) wraps fileObj in ControlStruct class
        # getNextCharacter is a method
        while getNextCharacter(ControlStruct(f)) != '':
            l = readnoncomment(f)
            r = re.match(r'\[(.*)\]\n', l)

            if r:
                tag = r.group(1)

                s = ''
                l = readnoncomment(f)
                while l not in ('\n', ''):
                    s += l
                    l = readnoncomment(f)

                syntax[tag] = s

        f.close()

    return syntax


def insertmenuitems(f, mname, current, prefix):
    """
  This function inserts menu items into the output html
  
  mname is the name of the menu file
  current is the 
  prefix is the
  f is the ControlStruct wrapper class that decorates the output file object
  """
    m = open(mname, 'rb')
    while getNextCharacter(ControlStruct(m)) != '':
        l = readnoncomment(m)
        l = l.strip()
        if l == '':
            continue

        r = re.match(r'\s*(.*?)\s*\[(.*)\]',
                     l)  # this matches: one or more spaces followed by

        if r:  # then we have a menu item.
            link = r.group(2)
            #print link
            # Don't use prefix if we have an absolute link.
            if '://' not in r.group(2):
                link = prefix + allreplace(link)

            # replace spaces with nbsps.
            # do do this, even though css would make it work - ie ignores.
            # only replace spaces that aren't in {{ blocks.
            in_quote = False
            menuitem = ""
            for group in re.split(r'({{|}})', r.group(1)):
                if in_quote:
                    if group == '}}':
                        in_quote = False
                        next
                    else:
                        menuitem += group
                else:
                    if group == '{{':
                        in_quote = True
                        next
                    else:
                        menuitem += br(re.sub(r'(?<!\\n) +', '~', group), f)

            if link[-len(current):] == current:
                hb(f.outf, f.conf['currentmenuitem'], link, menuitem)
            else:
                hb(f.outf, f.conf['menuitem'], link, menuitem)

        else:  # menu category.
            hb(f.outf, f.conf['menucategory'], br(l, f))

    m.close()


def out(f, s):
    f.write(s)


def hb(f, tag, content1, content2=None):
    """Writes out a halfblock (hb)."""

    if content1 is None:
        content1 = ""

    if content2 is None:
        out(f, re.sub(r'\|', content1, tag))
    else:
        r = re.sub(r'\|1', content1, tag)
        r = re.sub(r'\|2', content2, r)
        out(f, r)


def getNextCharacter(f, ditchcomments=True):
    """Peeks at next character in the file."""
    # Should only be used to look at the first character of a new line.
    c = f.inf.read(1)  # reads first charactor of input file
    if c:  # only undo forward movement if we're not at the end.
        #
        # if character c is comment symbol and ditchComments is True, read next line with getNextLine(f)
        if ditchcomments and c == '#':
            l = getNextLine(f)  # reads next line
            if doincludes(
                    f, l
            ):  # doincludes(fileObj, line) is a method described below. Checks if user is including external file
                return "#"

        if c in ' \t':
            return getNextCharacter(f)

        if c == '\\':
            c += getNextCharacter(f)

        f.inf.seek(-1, 1)
    elif f.otherfiles:
        f.nextfile()
        return getNextCharacter(f, ditchcomments)

    return c


def doincludes(f, l):
    """
  Input: f = fileObject, l = line
  This method includes jemdoc file that is included using the syntax include{filename} or includeraw{filename}
  includeraw does not jemdoc parsing
  include parses with jemdoc rules
  
  Returns boolean on completion, False if nothing has been included
  """
    ir = 'includeraw{'
    i = 'include{'
    if l.startswith(ir):  # if line starts with includeraw{
        nf = open(l[len(ir):-2], 'rb')
        rawHTML = nf.read()
        f.outf.write(
            rawHTML)  # write the contents of includeraw to the outputfile
        nf.close()
    elif l.startswith(i):  # if line starts with include{
        f.pushfile(
            l[len(i):-2]
        )  # switches which file is currently being parsed. makes the include{file} the one to be parsed
    else:
        return False

    return True


def getNextLine(f, withcount=False, codemode=False):
    """Get input file line."""
    """
  This method extracts the next line from the current file. If that file has no extra lines
  It switches to the next file
  """
    s = f.inf.readline()
    if not s and f.otherfiles:  # if current file has no lines left and there are other files
        f.nextfile()  # switch files
        return getNextLine(f, withcount, codemode)  # perform same algorithm on it

    f.linenum += 1

    if not codemode:
        # remove any special characters - assume they were checked by getNextCharacter()
        # before we got here.
        # remove any trailing comments.
        s = s.lstrip(' \t')
        s = re.sub(r'\s*(?<!\\)#.*', '', s)

    if withcount:
        if s[0] == '.':
            m = r'\.'
        else:
            m = s[0]

        r = re.match('(%s+) ' % m, s)
        if not r:
            raise SyntaxError("couldn't handle the jandal (code 12039) on line%d" % f.linenum)

        if not codemode:
            s = s.lstrip('-.=:')

        return (s, len(r.group(1)))
    else:
        if not codemode:
            s = s.lstrip('-.=:')

        return s


def nextParagraph(f, withcount=False, eatblanks=True):
    """Gets the next paragraph from the input file."""
    # New paragraph markers signalled by characters in following tuple.
    if withcount:
        (s, c) = getNextLine(f, withcount)
    else:
        s = getNextLine(f)

    while getNextCharacter(f) not in ('\n', '-', '.', ':', '', '=', '~', '{', '\\(', '\\)'):
        s += getNextLine(f)

    while eatblanks and getNextCharacter(f) == '\n':
        getNextLine(f)  # burn blank line.

    # in both cases, ditch the trailing \n.
    if withcount:
        return (s[:-1], c)
    else:
        return s[:-1]


def quote(s):
    """
  Input: string s
   This method escapes wildcards and reserved characters in the string s and returns it
   
  Return: escaped string s
  """
    return re.sub(r"""[\\*/+"'<>&$%\.~[\]-]""", r'\\\g<0>', s)


def replacequoted(b):
    """Quotes {{raw html}} sections."""
    """This method extracts the html within {{html text}} and returns that html text. HTML text will be escaped/sanitized as well"""
    r = re.compile(r'\{\{(.*?)\}\}', re.M + re.S)
    m = r.search(b)
    while m:
        qb = quote(m.group(1))

        b = b[:m.start()] + qb + b[m.end():]

        m = r.search(b, m.start())

    return b


def replacepercents(b):
    # replace %sections% as +{{sections}}+. Do not replace if within a link.

    r = re.compile(r'(?<!\\)%(.*?)(?<!\\)%', re.M + re.S)
    m = r.search(b)
    while m:
        #qb = '+' + quote(m.group(1)) + '+'
        a = re.sub(r'\[', r'BSNOTLINKLEFT12039XX', m.group(1))
        a = re.sub(r'\]', r'BSNOTLINKRIGHT12039XX', a)
        qb = '+{{' + a + '}}+'

        b = b[:m.start()] + qb + b[m.end():]

        m = r.search(b, m.start())

    return b

def replaceimages(b):
    # works with [img{width}{height}{alttext} location caption].
    r = re.compile(r'(?<!\\)\[img(\.[A-Za-z0-9]+)?(#[A-Za-z0-9]+)?((?:\{.*?\}){,3})\s(.*?)(?:\s(.*?))?(?<!\\)\]', re.M + re.S)
    m = r.search(b)
    s = re.compile(r'{(.*?)}', re.M + re.S)
    while m:
        imgClass = m.group(1)
        imgId = m.group(2)
        if imgClass:
            imgClass = imgClass[1:]

        if imgId:
            imgId = imgId[1:]

        m1 = list(s.findall(m.group(3)))
        m1 += [''] * (3 - len(m1))

        bits = []
        link = m.group(4).strip()
        bits.append(r'src=\"%s\"' % quote(link))

        if m1[0]:
            if m1[0].isdigit():
                s = m1[0] + 'px'
            else:
                s = m1[0]
            bits.append(r'width=\"%s\"' % quote(s))
        if m1[1]:
            if m1[1].isdigit():
                s = m1[1] + 'px'
            else:
                s = m1[1]
            bits.append(r'height=\"%s\"' % quote(s))
        if m1[2]:
            bits.append(r'alt=\"%s\"' % quote(m1[2]))
        else:
            bits.append(r'alt=\"\"')

        b = b[:m.start()] + r'<figure><img class="img-fluid %s" id="%s" %s /></figure>' % (imgClass, imgId, " ".join(bits) + b[m.end():])

        m = r.search(b, m.start())
    return b


def replacelinks(b):
    """
  This method replaces all links written in the meta language and produces the corresponding HTML anchor tag
  with the href attribute appropriately populated
  """
  # TODO: add class and id here
    # works with [link.html new link style].
    r = re.compile(r'(?<!\\)\[(.*?)(?:\s(.*?))?(?<!\\)\](\.[A-Za-z0-9]+)?(#[A-Za-z0-9]+)?', re.M + re.S)
    m = r.search(b)

    while m:
        m1 = m.group(1).strip()
        anchorClass = m.group(3).strip() if m.group(3) else None
        anchorId = m.group(4).strip() if m.group(4) else None

        if anchorClass:
            anchorClass = anchorClass[1:]
        
        if anchorId:
            anchorId = anchorId[1:]

        if '@' in m1 and not m1.startswith('mailto:') and not m1.startswith('http://'):
            link = 'mailto:' + m1
        else:
            link = m1

        # first unquote any hashes (e.g. for in-page links).
        link = re.sub(r'\\#', '#', link)

        # remove any +{{ or }}+ links.
        link = re.sub(r'(\+\{\{|\}\}\+)', r'%', link)

        link = quote(link)

        if m.group(2):
            linkname = m.group(2).strip()
        else:
            # remove any mailto before labelling.
            linkname = re.sub('^mailto:', '', link)

        aId = ''
        if anchorId:
            aId = 'id="%s"' % anchorId

        aClass = ''
        if anchorClass:
            aClass = 'class="%s"' % anchorClass


        b = b[:m.start()] + r'<a href=\"%s\" %s %s>%s<\/a>' % (link, aId, aClass, linkname) + b[m.end():]

        m = r.search(b, m.start())

    return b


def br(b, f, tableblock=False):
    """Does simple text replacements on a block of text. ('block replacements')"""

    # Deal with environment variables (say, for Michael Grant).
    r = re.compile(r"!\$(\w{2,})\$!", re.M + re.S)

    for m in r.findall(b):
        repl = os.environ.get(m)
        if repl == None:
            b = re.sub("!\$%s\$!" % m, 'FAILED_MATCH_' + m, b)
        else:
            b = re.sub("!\$%s\$!" % m, repl, b)

    b = re.sub(r'\\\\', 'jemLITerl33talBS', b)

    # Deal with {{html embedding}}.
    b = replacequoted(b)

    b = allreplace(b)

    b = b.lstrip('-. \t')  # remove leading spaces, tabs, dashes, dots.
    b = replaceimages(b)  # jem not sure if this is still used.

    # Slightly nasty hackery in this next bit.
    b = replacepercents(b)
    b = replacelinks(b)
    b = re.sub(r'BSNOTLINKLEFT12039XX', r'[', b)
    b = re.sub(r'BSNOTLINKRIGHT12039XX', r']', b)
    b = replacequoted(b)

    # Deal with /italics/ first because the '/' in other tags would otherwise
    # interfere.
    r = re.compile(r'(?<!\\)/(.*?)(?<!\\)/', re.M + re.S)
    b = re.sub(r, r'<em>\1</em>', b)

    # Deal with *bold*.
    r = re.compile(r'(?<!\\)\*(.*?)(?<!\\)\*', re.M + re.S)
    b = re.sub(r, r'<strong>\1</strong>', b)

    # Deal with +monospace+.
    r = re.compile(r'(?<!\\)\+(.*?)(?<!\\)\+', re.M + re.S)
    b = re.sub(r, r'<tt>\1</tt>', b)

    # Deal with "double quotes".
    r = re.compile(r'(?<!\\)"(.*?)(?<!\\)"', re.M + re.S)
    b = re.sub(r, r'&ldquo;\1&rdquo;', b)

    # Deal with left quote `.
    r = re.compile(r"(?<!\\)`", re.M + re.S)
    b = re.sub(r, r'&lsquo;', b)

    # Deal with apostrophe '.
    # Add an assertion that the next character's not a letter, to deal with
    # apostrophes properly.
    r = re.compile(r"(?<!\\)'(?![a-zA-Z])", re.M + re.S)
    b = re.sub(r, r'&rsquo;', b)

    # Deal with em dash ---.
    r = re.compile(r"(?<!\\)---", re.M + re.S)
    b = re.sub(r, r'&#8201;&mdash;&#8201;', b)

    # Deal with en dash --.
    r = re.compile(r"(?<!\\)--", re.M + re.S)
    b = re.sub(r, r'&ndash;', b)

    # Deal with ellipsis ....
    r = re.compile(r"(?<!\\)\.\.\.", re.M + re.S)
    b = re.sub(r, r'&hellip;', b)

    # Deal with non-breaking space ~.
    r = re.compile(r"(?<!\\)~", re.M + re.S)
    b = re.sub(r, r'&nbsp;', b)

    # Deal with registered trademark \R.
    r = re.compile(r"(?<!\\)\\R", re.M + re.S)
    b = re.sub(r, r'&reg;', b)

    # Deal with copyright \C.
    r = re.compile(r"(?<!\\)\\C", re.M + re.S)
    b = re.sub(r, r'&copy;', b)

    # Deal with middot \M.
    r = re.compile(r"(?<!\\)\\M", re.M + re.S)
    b = re.sub(r, r'&middot;', b)

    # Deal with line break.
    r = re.compile(r"(?<!\\)\\n", re.M + re.S)
    b = re.sub(r, r'<br />', b)

    # Deal with paragraph break. Caution! Should only use when we're already in
    # a paragraph.
    r = re.compile(r"(?<!\\)\\p", re.M + re.S)
    b = re.sub(r, r'</p><p>', b)

    if tableblock:
        # Deal with ||, meaning </td></tr><tr><td>
        r = re.compile(r"(?<!\\)\|\|", re.M + re.S)
        f.tablecol = 2
        bcopy = b
        b = ""
        r2 = re.compile(r"(?<!\\)\|", re.M + re.S)
        for l in bcopy.splitlines():
            f.tablerow += 1
            l = re.sub(r, r'</td></tr>\n<tr class="r%d"><td class="c1">' \
                  % f.tablerow, l)

            l2 = ''
            col = 2
            r2s = r2.split(l)
            for x in r2s[:-1]:
                l2 += x + ('</td><td class="c%d">' % col)
                col += 1
            l2 += r2s[-1]

            b += l2

    # Second to last, remove any remaining quoting backslashes.
    b = re.sub(r'\\(?!\\)', '', b)

    # Deal with literal backspaces.
    b = re.sub('jemLITerl33talBS', r'\\', b)

    # Also fix up DOUBLEOPEN and DOUBLECLOSEBRACES.
    b = re.sub('DOUBLEOPENBRACE', '{{', b)
    b = re.sub('DOUBLECLOSEBRACE', '}}', b)

    return b


def allreplace(b):
    """Replacements that should be done on everything."""
    """
  Input: string b
  Return: escaped ampersand, less than and greater than symbols
  """
    # escaping/replacing & with amp
    r = re.compile(r"(?<!\\)&", re.M + re.S)
    b = re.sub(r, r'&amp;', b)

    # escaping/replacing greater than
    r = re.compile(r"(?<!\\)>", re.M + re.S)
    b = re.sub(r, r'&gt;', b)

    # escaping/replacing less than
    r = re.compile(r"(?<!\\)<", re.M + re.S)
    b = re.sub(r, r'&lt;', b)

    return b


def pyint(f, l):
    l = l.rstrip()
    l = allreplace(l)

    r = re.compile(r'(#.*)')
    l = r.sub(r'<span class = "comment">\1</span>', l)

    if l.startswith('&gt;&gt;&gt;'):
        hb(f, '<span class="pycommand">|</span>\n', l)
    else:
        out(f, l + '\n')


def putbsbs(l):
    for i in range(len(l)):
        l[i] = '\\b' + l[i] + '\\b'

    return l


def gethl(lang):
    # disable comments by default, by choosing unlikely regex.
    d = {'strings': False}
    if lang in ('py', 'python'):
        d['statement'] = [
            'break', 'continue', 'del', 'except', 'exec', 'finally', 'pass',
            'print', 'raise', 'return', 'try', 'with', 'global', 'assert',
            'lambda', 'yield', 'def', 'class', 'for', 'while', 'if', 'elif',
            'else', 'import', 'from', 'as', 'assert'
        ]
        d['builtin'] = [
            'True', 'False', 'set', 'open', 'frozenset', 'enumerate', 'object',
            'hasattr', 'getattr', 'filter', 'eval', 'zip', 'vars', 'unicode',
            'type', 'str', 'repr', 'round', 'range', 'and', 'in', 'is', 'not',
            'or'
        ]
        d['special'] = [
            'cols', 'optvar', 'param', 'problem', 'norm2', 'norm1', 'value',
            'minimize', 'maximize', 'rows', 'rand', 'randn', 'printval',
            'matrix'
        ]
        d['error'] = [
            '\w*Error',
        ]
        d['commentuntilend'] = '#'
        d['strings'] = True
    elif lang in ['c', 'c++', 'cpp']:
        d['statement'] = ['if', 'else', 'printf', 'return', 'for']
        d['builtin'] = [
            'static', 'typedef', 'int', 'float', 'double', 'void', 'clock_t',
            'struct', 'long', 'extern', 'char'
        ]
        d['operator'] = [
            '#include.*', '#define', '@pyval{', '}@', '@pyif{', '@py{'
        ]
        d['error'] = [
            '\w*Error',
        ]
        d['commentuntilend'] = ['//', '/*', ' * ', '*/']
    elif lang in ('rb', 'ruby'):
        d['statement'] = putbsbs([
            'while', 'until', 'unless', 'if', 'elsif', 'when', 'then', 'else',
            'end', 'begin', 'rescue', 'class', 'def'
        ])
        d['operator'] = putbsbs(['and', 'not', 'or'])
        d['builtin'] = putbsbs(['true', 'false', 'require', 'warn'])
        d['special'] = putbsbs(['IO'])
        d['error'] = putbsbs([
            '\w*Error',
        ])
        d['commentuntilend'] = '#'
        d['strings'] = True
        d['strings'] = True
        if lang in ['c++', 'cpp']:
            d['builtin'] += ['bool', 'virtual']
            d['statement'] += ['new', 'delete']
            d['operator'] += ['&lt;&lt;', '&gt;&gt;']
            d['special'] = [
                'public', 'private', 'protected', 'template', 'ASSERT'
            ]
    elif lang == 'sh':
        d['statement'] = [
            'cd',
            'ls',
            'sudo',
            'cat',
            'alias',
            'for',
            'do',
            'done',
            'in',
        ]
        d['operator'] = [
            '&gt;', r'\\', r'\|', ';', '2&gt;', 'monolith&gt;', 'kiwi&gt;',
            'ant&gt;', 'kakapo&gt;', 'client&gt;'
        ]
        d['builtin'] = putbsbs([
            'gem', 'gcc', 'python', 'curl', 'wget', 'ssh', 'latex', 'find',
            'sed', 'gs', 'grep', 'tee', 'gzip', 'killall', 'echo', 'touch',
            'ifconfig', 'git', '(?<!\.)tar(?!\.)'
        ])
        d['commentuntilend'] = '#'
        d['strings'] = True
    elif lang == 'matlab':
        d['statement'] = putbsbs([
            'max', 'min', 'find', 'rand', 'cumsum', 'randn', 'help', 'error',
            'if', 'end', 'for'
        ])
        d['operator'] = ['&gt;', 'ans =', '>>', '~', '\.\.\.']
        d['builtin'] = putbsbs(['csolve'])
        d['commentuntilend'] = '%'
        d['strings'] = True
    elif lang == 'commented':
        d['commentuntilend'] = '#'

    # Add bsbs (whatever those are).
    for x in ['statement', 'builtin', 'special', 'error']:
        if x in d:
            d[x] = putbsbs(d[x])

    return d


def language(f, l, hl):
    l = l.rstrip()
    l = allreplace(l)
    # handle strings.
    if hl['strings']:
        r = re.compile(r'(".*?")')
        l = r.sub(r'<span CLCLclass="string">\1</span>', l)
        r = re.compile(r"('.*?')")
        l = r.sub(r'<span CLCLclass="string">\1</span>', l)

    if 'statement' in hl:
        r = re.compile('(' + '|'.join(hl['statement']) + ')')
        l = r.sub(r'<span class="statement">\1</span>', l)

    if 'operator' in hl:
        r = re.compile('(' + '|'.join(hl['operator']) + ')')
        l = r.sub(r'<span class="operator">\1</span>', l)

    if 'builtin' in hl:
        r = re.compile('(' + '|'.join(hl['builtin']) + ')')
        l = r.sub(r'<span class="builtin">\1</span>', l)

    if 'special' in hl:
        r = re.compile('(' + '|'.join(hl['special']) + ')')
        l = r.sub(r'<span class="special">\1</span>', l)

    if 'error' in hl:
        r = re.compile('(' + '|'.join(hl['error']) + ')')
        l = r.sub(r'<span class="error">\1</span>', l)

    l = re.sub('CLCLclass', 'class', l)

    if 'commentuntilend' in hl:
        cue = hl['commentuntilend']
        if isinstance(cue, (list, tuple)):
            for x in cue:
                if l.strip().startswith(x):
                    hb(f, '<span class="comment">|</span>\n', allreplace(l))
                    return
                if '//' in cue:  # Handle this separately.
                    r = re.compile(r'\/\/.*')
                    l = r.sub(r'<span class="comment">\g<0></span>', l)
        elif cue == '#':  # Handle this separately.
            r = re.compile(r'#.*')
            l = r.sub(r'<span class="comment">\g<0></span>', l)
        elif cue == '%':  # Handle this separately.
            r = re.compile(r'%.*')
            l = r.sub(r'<span class="comment">\g<0></span>', l)
        elif l.strip().startswith(cue):
            hb(f, '<span class="comment">|</span>\n', allreplace(l))
            return

    out(f, l + '\n')

def dashlist(f, ordered=False):
    level = 0

    if ordered:
        char = '.'
        ul = 'ol'
    else:
        char = '-'
        ul = 'ul'

    while getNextCharacter(f) == char:
        (s, newlevel) = nextParagraph(f, True, False)

        # first adjust list number as appropriate.
        if newlevel > level:
            for i in range(newlevel - level):
                if newlevel > 1:
                    out(f.outf, '\n')
                out(f.outf, '<%s>\n<li>' % ul)
        elif newlevel < level:
            out(f.outf, '\n</li>')
            for i in range(level - newlevel):
                #out(f.outf, '</li>\n</%s>\n</li><li>' % ul)
                # demote means place '</ul></li>' in the file.
                out(f.outf, '</%s>\n</li>' % ul)
            #out(f.outf, '\n<li>')
            out(f.outf, '\n<li>')
        else:
            # same level, make a new list item.
            out(f.outf, '\n</li>\n<li>')

        out(f.outf, '<p>' + br(s, f) + '</p>')
        level = newlevel

    for i in range(level):
        out(f.outf, '\n</li>\n</%s>\n' % ul)


def colonlist(f):
    out(f.outf, '<dl>\n')
    while getNextCharacter(f) == ':':
        s = nextParagraph(f, eatblanks=False)
        r = re.compile(r'\s*{(.*?)(?<!\\)}(.*)', re.M + re.S)
        g = re.match(r, s)

        if not g or len(g.groups()) != 2:
            raise SyntaxError("couldn't handle the jandal (invalid deflist "
                              "format) on line %d" % f.linenum)
        # split into definition / non-definition part.
        defpart = g.group(1)
        rest = g.group(2)

        hb(f.outf, '<dt>|</dt>\n', br(defpart, f))
        hb(f.outf, '<dd><p>|</p></dd>\n', br(rest, f))

    out(f.outf, '</dl>\n')


def codeblock(f, g):
    # TODO: give users the ability to add class and id
    if g[1] == 'raw':
        raw = True
        ext_prog = None
    elif g[0] == 'filter_through':
        # Filter through external program.
        raw = False
        ext_prog = g[1]
        buff = ""
    else:
        ext_prog = None
        raw = False
        out(f.outf, f.conf['codeblock'])
        if g[0]:
            hb(f.outf, f.conf['blocktitle'], g[0])
        if g[1] == 'jemdoc':
            out(f.outf, f.conf['codeblockcontenttt'])
        else:
            out(f.outf, f.conf['codeblockcontent'])

    # Now we are handling code.
    # Handle \~ and ~ differently.
    stringmode = False
    while 1:  # wait for EOF.
        l = getNextLine(f, codemode=True)
        if not l:
            break
        elif l.startswith('~'):
            break
        elif l.startswith('\\~'):
            l = l[1:]
        elif l.startswith('\\{'):
            l = l[1:]
        elif ext_prog:
            buff += l
            continue
        elif stringmode:
            if l.rstrip().endswith('"""'):
                out(f.outf, l + '</span>')
                stringmode = False
            else:
                out(f.outf, l)
            continue

        # jem revise pyint out of the picture.
        if g[1] == 'pyint':
            pyint(f.outf, l)
        else:
            if raw:
                out(f.outf, l)
            elif g[1] == 'jemdoc':
                # doing this more nicely needs python 2.5.
                for x in ('#', '~', '>>>', '\~', '{'):
                    if str(l).lstrip().startswith(x):
                        out(f.outf, '</tt><pre class="tthl">')
                        out(f.outf, l + '</pre><tt class="tthl">')
                        break
                else:
                    for x in (':', '.', '-'):
                        if str(l).lstrip().startswith(x):
                            out(f.outf, '<br />' + prependnbsps(l))
                            break
                    else:
                        if str(l).lstrip().startswith('='):
                            out(f.outf, prependnbsps(l) + '<br />')
                        else:
                            out(f.outf, l)
            else:
                if l.startswith('\\#include{') or l.startswith(
                        '\\#includeraw{'):
                    out(f.outf, l[1:])
                elif l.startswith('#') and doincludes(f, l[1:]):
                    continue
                elif g[1] in ('python', 'py') and l.strip().startswith('"""'):
                    out(f.outf, '<span class="string">' + l)
                    stringmode = True
                else:
                    language(f.outf, l, gethl(g[1]))

    if raw:
        return
    elif ext_prog:
        print 'filtering through %s...' % ext_prog

        output, _ = subprocess.Popen(ext_prog, shell=True, stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE).communicate(buff)
        out(f.outf, output)
    else:
        if g[1] == 'jemdoc':
            out(f.outf, f.conf['codeblockendtt'])
        else:
            out(f.outf, f.conf['codeblockend'])


def prependnbsps(l):
    g = re.search('(^ *)(.*)', l).groups()
    return g[0].replace(' ', '&nbsp;') + g[1]


def inserttitle(f, t):
    # TODO: give users the ability to add id and class
    if t is not None:
        hb(f.outf, f.conf['doctitle'], t)

        # Look for a subtitle.
        if getNextCharacter(f) != '\n':
            hb(f.outf, f.conf['subtitle'], br(nextParagraph(f), f))

        hb(f.outf, f.conf['doctitleend'], t)


def insertnavbaritems(f, mname, current, prefix):
    """
  This function inserts navbar items into the navbar html
  
  mname is the name of the menu file
  current is the 
  prefix is the
  """
    m = open(mname, 'rb')
    while getNextCharacter(ControlStruct(m)) != '':
        l = readnoncomment(m)
        l = l.strip()
        if l == '':
            continue

        r = re.match(r'\s*(.*?)\s*\[(.*)\]',
                     l)  # this matches: one or more spaces followed by

        if r:  # then we have a menu item.
            link = r.group(2)
            # Don't use prefix if we have an absolute link.
            if '://' not in r.group(2):
                link = prefix + allreplace(link)

            # replace spaces with nbsps.
            # do do this, even though css would make it work - ie ignores.
            # only replace spaces that aren't in {{ blocks.
            in_quote = False
            navitem = ""
            for group in re.split(r'({{|}})', r.group(1)):
                if in_quote:
                    if group == '}}':
                        in_quote = False
                        next
                    else:
                        navitem += group
                else:
                    if group == '{{':
                        in_quote = True
                        next
                    else:
                        navitem += br(re.sub(r'(?<!\\n) +', '~', group), f)

            if link[-len(current):] == current:
                hb(f.outf, f.conf['currentnavitem'], link, navitem)
            else:
                hb(f.outf, f.conf['navitem'], link, navitem)
    m.close()

def procfile(f, cliparser):
    f.linenum = 0

    menu = {}
    navbar = {}
    # convert these to a dictionary.
    showfooter = True
    showsourcelink = False
    showlastupdated = True
    showlastupdatedtime = True
    nodefaultcss = False
    nav = False
    fwtitle = False
    css = []
    js = []
    title = None

    while getNextCharacter(f, False) == '#':
        l = f.inf.readline()

        f.linenum += 1
        if doincludes(f, l[1:]):
            continue
        if l.startswith('# jemdoc:'):
            l = l[len('# jemdoc:'):]
            a = l.split(',')
            # jem only handle one argument for now.
            for b in a:
                b = b.strip()
                if b.startswith('menu'):
                    sidemenu = True
                    r = re.compile(r'(\.[A-Za-z0-9]+)?(#[A-Za-z0-9]+)?(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    g = re.findall(r, b)
                    if len(g) > 3 or len(g) < 2:
                        raise SyntaxError('sidemenu error on line %d' % f.linenum)
                    if len(g) == 2:
                        menu = {'file': f, 'menuname': g[0][2], 'page': g[1][2], 'preface': ''}
                        # menu = (f, g[0][2], g[1][2], '')
                    else:
                        menu = {'file': f, 'menuname': g[0][2], 'page': g[1][2], 'preface': g[2][2]}
                        # menu = (f, g[0][2], g[1][2], g[2][2])
                        # 'class': g[0][0], 'id': g[0][1]
                    if g[0][0]:
                        menu['class'] = g[0][0]
                    if g[0][1]:
                        menu['id'] = g[0][1]

                # TODO: setup navbar like menu
                elif b.startswith('nav'):
                    # navbar
                    nav = True
                    r = re.compile(r'(\.[A-Za-z0-9]+)?(#[A-Za-z0-9]+)?(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    g = re.findall(r, b)
                    if len(g) > 3 or len(g) < 2:
                        raise SyntaxError('navbar error on line %d' % f.linenum)
                    if len(g) == 2:
                        navbar = {'file': f, 'navname': g[0][2], 'page': g[1][2], 'preface': ''}
                        # navbar = (f, g[0], g[1], '')
                    else:
                        navbar = {'file': f, 'navname': g[0][2], 'page': g[1][2], 'preface': g[2][2]}
                        # navbar = (f, g[0], g[1], g[2])
                    if g[0][0]:
                        navbar['class'] = g[0][0]
                    if g[0][1]:
                        navbar['id'] = g[0][1]

                elif b.startswith('nofooter'):
                    showfooter = False

                elif b.startswith('nodate'):
                    showlastupdated = False

                elif b.startswith('notime'):
                    showlastupdatedtime = False

                elif b.startswith('fwtitle'):
                    fwtitle = True
                # elif b.startswith('footerid'): # add footer id and class
                #     r = re.compile(r'')
                elif b.startswith('showsource'):
                    showsourcelink = True

                elif b.startswith('nodefaultcss'):
                    nodefaultcss = True

                elif b.startswith('addcss'):
                    r = re.compile(r'(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    cssFiles = re.findall(r, b)  
                    css.extend(cssFiles)

                elif b.startswith('addjs'):
                    r = re.compile(r'(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    jsFiles = re.findall(r, b)
                    js.extend(jsFiles)

                elif b.startswith('addtex'):
                    r = re.compile(r'(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    textLines = re.findall(r, b) 
                    f.texlines += textLines

                elif b.startswith('analytics'):
                    r = re.compile(r'(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    analytics = re.findall(r, b)[0]  
                    f.analytics = analytics

                elif b.startswith('title'):
                    r = re.compile(r'(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    g = re.findall(r, b)
                    if len(g) != 1:
                        raise SyntaxError('addtitle error on line %d' %f.linenum)
                    title = g[0]

    # Get the file started with the firstbit.
    # writing the first bit to the output file
    out(f.outf, f.conf['firstbit'])

    # writing default css link tag to output file
    if not nodefaultcss:
        outdir = f.getOutputDir()
        Style.downloadDefaultCSS(outdir)
        out(f.outf, f.conf['defaultcss'])

    # Add per-file css lines here.
    extension = '.' + cliparser.getCssStyle()
    invalidCSS = []
    for cssFile in css:
        # compile with corresponding css engine
        outCssfile = os.path.splitext(cssFile)[0] + '.css'
        cssEngine = cliparser.getCssEngine()
        if cssFile.endswith(extension) and extension != ".css":
            compiled = CommandLineParser.compileToCss(cssEngine, cssFile, outCssfile)
            if not compiled:
                invalidCSS.append(cssFile)
        elif '.css' not in cssFile:
            cssFile += '.css'

    # writing the CSS html link tags to the output file
    for cssFile in css:
        if cssFile not in invalidCSS:
            hb(f.outf, f.conf['specificcss'], cssFile)

    # writing the JS html script tags to the output file
    for jsFile in js:
        hb(f.outf, f.conf['specificjs'], jsFile)

    # Look for a title.
    if getNextCharacter(f) == '=':  # don't check exact number f.outf '=' here jem.
        t = br(getNextLine(f), f)[:-1]
        if title is None:
            title = re.sub(' *(<br />)|(&nbsp;) *', ' ', t)
    else:
        t = None

    #if title:
    hb(f.outf, f.conf['windowtitle'], title)

    out(f.outf, f.conf['bodystart'])

    if f.analytics:
        hb(f.outf, f.conf['analytics'], f.analytics)

    if fwtitle:
        out(f.outf, f.conf['fwtitlestart'])
        inserttitle(f, t)
        out(f.outf, f.conf['fwtitleend'])

    # navbar
    if navbar:
        # navbar = {'file': f, 'class': g[0][0], 'id': g[0][1], 'menuname': g[0][2], 'page': g[1][2], 'preface': g[2][2]} 
        navstart = f.conf['navstart'] % (navbar.get('class', ''), navbar.get('id', 'navbar-menu'))
        out(f.outf, navstart)
        # navbar = (f, 'menuname', 'page', 'preface')
        navData = ( navbar['file'], navbar['menuname'], navbar['page'], navbar['preface'] ) 
        insertnavbaritems(*navbar)
        out(f.outf, f.conf['navend'])
    else:
        out(f.outf, f.conf['nonav'])

    if menu:
        # menu = {'file': f, 'class': g[0][0], 'id': g[0][1], 'menuname': g[0][2], 'page': g[1][2], 'preface': g[2][2]}
        menustart = f.conf['menustart'] % (menu.get('class', ''), menu.get('id', 'sidemenu'))
        # print menustart
        out(f.outf, menustart )
        # menu = (f, 'menuname', 'page', 'preface')
        menuData = ( menu['file'], menu['menuname'], menu['page'], menu['preface'] ) 
        insertmenuitems(*menuData)
        out(f.outf, f.conf['menuend'])
    else:
        out(f.outf, f.conf['nomenu'])

    if not fwtitle:
        inserttitle(f, t)

    infoblock = False
    imgblock = False
    tableblock = False
    while 1:  # wait for EOF.
        p = getNextCharacter(f)

        if p == '':
            break

        # look for lists.
        elif p == '-':
            dashlist(f, False)

        elif p == '.':
            dashlist(f, True)

        elif p == ':':
            colonlist(f)

        # look for titles.
        elif p == '=':
            (s, count) = getNextLine(f, True)
            # trim trailing \n.
            s = s[:-1]
            hb(f.outf, '<h%d>|</h%d>\n' % (count, count), br(s, f))

        # look for comments.
        elif p == '#':
            l = getNextLine(f)

        elif p == '\n':
            getNextLine(f)

        # look for blocks.
        elif p == '~':
            getNextLine(f)
            # TODO: add class and id here
            # add class and id here
            if infoblock:
                out(f.outf, f.conf['infoblockend'])
                infoblock = False
                getNextLine(f)
                continue
            elif imgblock:
                out(f.outf, '</figure>\n')
                imgblock = False
                getNextLine(f)
                continue
            elif tableblock:
                out(f.outf, '</td></tr></table>\n')
                tableblock = False
                getNextLine(f)
                continue
            else:
                if getNextCharacter(f) == '{':
                    l = allreplace(getNextLine(f))
                    r = re.compile(r'(?<!\\){(.*?)(?<!\\)}', re.M + re.S)
                    g = re.findall(r, l)
                else:
                    g = []

                # process jemdoc markup in titles.
                if len(g) >= 1:
                    g[0] = br(g[0], f)

                if len(g) in (0, 1):  # info block.
                    out(f.outf, f.conf['infoblock'])
                    infoblock = True

                    if len(g) == 1:  # info block.
                        hb(f.outf, f.conf['blocktitle'], g[0])
                    out(f.outf, f.conf['infoblockcontent'])

                elif len(g) >= 2 and g[1] == 'table':
                    # handles
                    # {title}{table}{name}
                    # one | two ||
                    # three | four ||
                    tableClass = None
                    tableSplit = g[1].split(".")
                    if len(tableSplit) > 1:
                        tableClass = tableSplit[1]
                    name = ''
                    if len(g) >= 3 and g[2]:
                        name += ' id="%s"' % g[2]
                    if tableClass:
                        name += ' class="%s"' %tableClass
                    out(f.outf, '<table%s>\n<tr class="r1"><td class="c1">' % name)
                    f.tablerow = 1
                    f.tablecol = 1

                    tableblock = True

                elif len(g) == 2:
                    codeblock(f, g)

                elif len(g) >= 4 and g[1] == 'img_left':
                    # handles
                    # {}{img_left}{source}{alttext}{width}{height}{linktarget}.
                    imgClass = None
                    imgId = None
                    imgClassSplit = g[1].split('.')
                    imgIdSplit = g[1].split('#')
                    if len(imgClassSplit) > 1:
                        imgClass = imgClassSplit[1]

                    if len(imgIdSplit) > 1:
                        imgId = imgIdSplit[1]

                    g += [''] * (7 - len(g))

                    if g[4].isdigit():
                        g[4] += 'px'

                    if g[5].isdigit():
                        g[5] += 'px'

                    out(f.outf, '<figure>\n')
                    if g[6]:
                        out(f.outf, '<a href="%s">' % g[6])
                    
                    htmlClass = ''
                    if imgClass:
                        htmlClass = 'class="%s"' %imgClass
                    
                    htmlId = ''
                    if imgId:
                        htmlId = 'id="%s"' %imgId

                    out(f.outf, '<img class="img-fluid" src="%s" %s %s' %(g[2], htmlClass, htmlId) )
                    out(f.outf, ' alt="%s"' % g[3])
                    if g[4]:
                        out(f.outf, ' width="%s"' % g[4])
                    if g[5]:
                        out(f.outf, ' height="%s"' % g[5])
                    out(f.outf, ' />')
                    if g[6]:
                        out(f.outf, '</a>')
                    imgblock = True
                
                elif len(g) >= 2 and 'video' in g[1]:
                    # {}{video}{width}{height}{source}{source}...
                    videoTag = '<video'
                    if 'c' in g[1]:
                        videoTag += ' controls'
                    
                    if 'a' in g[1]:
                        videoTag += ' autoplay'
                    
                    if 'l' in g[1]:
                        videoTag += ' loop'

                    start = 2
                    if g[2].isdigit():
                        g[2] += 'px'
                        videoTag += ' width="%s"' %g[2]
                        start = 3

                    if g[3].isdigit():
                        g[3] += 'px'
                        videoTag += ' height="%s"' %g[3]
                        start = 4

                    videoTag += '>\n'
                    out(f.outf, videoTag)

                    for i in range(start, len(g)):
                        if g[i]:
                            source = g[i]
                            ext = os.path.splitext(g[i])[1]
                            srcTag = '<source src=%s type=video/%s />\n' %(source, ext)
                            out(f.outf, srcTag)
                        else:
                            # user did not provide source and so we consider it to be the end of it
                            break
                    out(f.outf, '<em>Sorry, your browser <strong>does not</strong> support the embedded videos.</em>')
                    out(f.outf, '</video>\n')

                # parse audio
                elif len(g) >= 2 and 'audio' in g[1]:
                    # {}{audio}{source}{source}...
                    classList = g[1].split('.')
                    idList = g[1].split('#')
                    audioClass = None
                    audioId = None

                    if len(classList) > 1:
                        audioClass = str(classList[1])
                    
                    if len(idList) > 1:
                        audioId = str(idList[1])

                    audioTag = '<audio'
                    if audioClass:
                        audioTag += ' %s' % audioClass

                    if audioId:
                        audioTag += ' %s' % audioId

                    if 'c' in g[1]:
                        audioTag += ' controls'
                    
                    if 'a' in g[1]:
                        audioTag += ' autoplay'
                    
                    if 'l' in g[1]:
                        audioTag += ' loop'
                    audioTag += '>\n'
                    out(f.outf, audioTag)

                    for i in range(2, len(g)):
                        if g[i]:
                            source = g[i]
                            ext = os.path.splitext(g[i])[1]
                            srcTag = '<source src=%s type=audio/%s />\n' %(source, ext)
                            out(f.outf, srcTag)
                        else: # user did not provide source and so we consider it to be the end of it
                            break
                    out(f.outf, '<em>Sorry, your browser <strong>does not</strong> support embedded audios.</em>')
                    out(f.outf, '</audio>\n')
                
                # parse forms
                elif len(g) >= 2 and g[1] == 'fs' or g[1] == 'FS':
                    # {}{fs}{action}{method}{name}
                    #g += [''] * (7 - len(g))
                    
                    classList = g[1].split('.')
                    idList = g[1].split('#')
                    formClass = None
                    formId = None

                    if len(classList) > 1:
                        formClass = str(classList[1])
                    
                    if len(idList) > 1:
                        formId = str(idList[1])

                    if not g[2]:
                        raisejandal("cannot create form without action attribute provided: \{fs\}\{action\}", f.linenum)
                    htmlForm = '<form action="%s"' %g[2]

                    if formId:
                        htmlForm += ' id="%s"' % formId
                    if formClass:
                        htmlForm += ' class="%s"' % formClass
                    
                    out(f.out, htmlForm)
                    
                    if g[3]:
                        out(f.outf, ' method="%s"') %g[3]
                    else:
                        out(f.outf, ' method="GET"')
 
                    if 'v' in g[1] or 'V' in g[1]:
                        out(f.outf, ' novalidate')
                    # introduced new form block variable for creation of forms
                    out(f.out, '>\n')

                elif len(g) == 2 and g[1].lower() == 'fe':
                    # {}{fe}
                    out(f.outf, '</form>\n')       
                elif len(g) >= 2 and g[1].lower() in ('it', 'ip', 'ie', 'ic', 'is'):
                    #{}{it}{name}{placeholder} = input type="text" name value
                    #{}{ip}{name}{placeholder} = input type="password" name
                    #{}{ie}{name}{placeholder} = input type="email" name
                    #{}{ic}{name}{value}{label} = input type="checkbox" name value 
                    #{}{is} =
                    g[1] = g[1].lower()
                    if g[1] != "is":
                        if not g[2]:
                            raisejandal("cannot create <input> tag without a 'name' attribute provided\nusage: \{inputtag\}\{name\}", f.linenum)
                        if g[1] == 'ic' and not g[3]:
                            raisejandal("cannot create <input type='checkbox> without a 'value' attrubute\nusage: \{inputtag\}\{name\}\{value\}")

                    if g[1] in ('it', 'ip', 'ie'):
                        name = str(g[2])
                        placeholder = None

                        if (g[3]):
                            placeholder = str(g[3])                        
                        inputType = 'text'
                        if g[1] == 'ip':
                            inputType = 'password'
                        elif g[1] == 'ie':
                            inputType = 'email'
                        out(f.outf, '<div class="form-group">\n')
                        out(f.outf, '<input type="%s" name="%s" placeholder="%s" class="form-control" />\n' %(inputType, name, placeholder))
                        out(f.outf, '</div>\n')
                    elif g[1] == 'is':
                        out(f.outf, '<button type="submit" class="btn btn-primary">Submit</button>\n')
                    elif g[1] == 'ic':
                        name = str(g[2])
                        inputType = 'checkbox'
                        value = str(g[3])
                        out(f.outf, '<div class="form-group">\n')
                        out(f.outf, '<input type="%s" name="%s" value="%s" class="form-control" />\n' %(inputType, name, value))
                        out(f.outf, '</div>\n')

                else:
                    raisejandal("couldn't handle block", f.linenum)

        else:
            s = br(nextParagraph(f), f, tableblock)
            if s:
                if tableblock:
                    hb(f.outf, '|\n', s)
                else:
                    hb(f.outf, '<p>|</p>\n', s)

    if menu:
        out(f.outf, f.conf['menulastbit'])
    else:
        out(f.outf, f.conf['nomenulastbit'])

    if showfooter and (showlastupdated or showsourcelink):
        out(f.outf, f.conf['footerstart'])
        if showlastupdated:
            if showlastupdatedtime:
                ts = '%Y-%m-%d %H:%M:%S %Z'
            else:
                ts = '%Y-%m-%d'
            s = time.strftime(ts, time.localtime(time.time()))
            hb(f.outf, f.conf['lastupdated'], s)
        if showsourcelink:
            hb(f.outf, f.conf['sourcelink'], f.inname)
        out(f.outf, f.conf['footerend'])

    out(f.outf, f.conf['bodyend'])

    if f.outf is not sys.stdout:
        # jem: close file here.
        # jem: XXX this is where you would intervene to do a fast open/close.
        f.outf.close()


# # TODO: regex matching bug for optional capture group


#!/usr/bin/env python2

import argparse, os, subprocess, sys, shutil, pkg_resources

from defaultconfig import DefaultConfig

PKG_NAME = "jemdoc"

class CommandLineParser(object):
    '''
    The CommandLineParser parses and processes command line arguments provided by the user.
    
    Args:

    Attributes:
        parser (ArgumentParser): this is the backbone for parsing cli arguments
        _cssEngine (str): represents the css preprocessor the user is using
    '''
    def __init__(self):
        self._standardConfig = DefaultConfig()
        self._DEFAULT_OUTDIR = "."
        self._version = "0.8.0"
        self._cssEngine = None
        self.parser = argparse.ArgumentParser(
            prog=PKG_NAME,
            description=
            "Produces html markup from a jemdoc SOURCEFILE.\n\nSee \033[96mhttp://jemdoc.jaboc.net/\033[00m for many more details."
        )
        self.parser.add_argument(
            "-s",
            "--show-config",
            dest="showConfig",
            help=
            "downloads the standard configuration 'standardconfig.txt' in the current directory",
            action="store_true")
        self.parser.add_argument("-v",
                                 "--version",
                                 dest="version",
                                 help="display Jemdoc version number",
                                 action="version",
                                 version="%(prog)s {}".format(self._version))
        self.parser.add_argument("-i",
                                 "--info",
                                 dest="info",
                                 help="display system information",
                                 action="store_true")
        self.parser.add_argument(
            "--css",
            dest="cssEngine",
            metavar="\033[94m<engine>\033[00m",
            help=
            "add stylesheet <engine> support [less|sass|scss] (defaults to plain css)",
            type=self._checkCssEngine,
            default="css")
        # self.parser.add_argument("-r",
        #                          "--react",
        #                          dest="useReact",
        #                          help="compile with ReactJS",
        #                          action="store_true")
        self.parser.add_argument("-c",
                                 "--config",
                                 dest="config",
                                 metavar="\033[94m<filename>\033[00m",
                                 help="add custom configuration file",
                                 action="store")
        self.parser.add_argument(
            "-d",
            "--dir",
            dest="outdir",
            metavar="\033[94m<dirname>\033[00m",
            help="output files in directory <dirname>", # name of ReactJS directory when compiling with ReactJS",
            type=self._checkOutputDir)
        self.parser.add_argument("input",
                                 help="relative path to input files",
                                 type=self._checkFileExists,
                                 nargs="*")
        self._args = self.parser.parse_args()
        # self.createStandardConfig()

    def getArgs(self):
        return self._args

    def getStandardConfigPath(self):
        return os.path.join(os.getcwd(), "configs", "standardconf.config")

    def getInputFiles(self):
        return self._args.input

    def _checkOutputDir(self, outdir):
        if os.path.isfile(outdir):
            raise argparse.ArgumentTypeError(
                "output directory name '{}' is not a directory, please give a directory!"
                .format(outdir))

        if os.path.exists(outdir):  # already exists
            prompt = "'{}' already exists. Would you like to override it (\033[92my\033[00m/\033[91mn\033[00m) ".format(
                outdir)
            response = raw_input(prompt).lower()
            while response not in ('y', 'yes', 'yeah', 'yah', 'n', 'no',
                                   'nah'):
                response = raw_input(prompt).lower()

            if response in ('y', 'yes', 'yeah', 'yah'):
                shutil.rmtree(outdir)
                os.mkdir(outdir)
            else:
                raise argparse.ArgumentTypeError(
                    "Duplicate directory name '{}'".format(outdir))
        else:
            os.mkdir(outdir)
        return outdir

    def _createOutputDir(self, dirname):
        try:
            os.mkdir(dirname)
        except FileExistsError as e:
            response = ""
            while response not in ("y", "yes", "no", "n"):
                response = raw_input("Would you like to overrwrite directory %s (\033[92my\033[00m/\033[91mn\033[00m) " % dirname).lower()
            if response in ("y", "yes"):
                shutil.rmtree(dirname)
                os.mkdir(dirname)
            else:
                raise FileExistsError("directory %s already exists" % dirname)

    def getOutputDir(self):
        if self._args.outdir:     
            return self._args.outdir
        else:
            return self._DEFAULT_OUTDIR

    def getUserConfig(self):
        if self._args.config:
            if not os.path.exists():
                raise FileNotFoundError("configuration file %s provided does not exist" % self._args.config)
        return self._args.config

    def showConfig(self):
        '''

        '''
        return self._args.showConfig

    def displayConfig(self):
        print CommandLineParser.getStandardConfig()
        sys.exit(1)

    def showInfo(self):
        return self._args.info

    def displayInfo(self):
        print 'Platform: ' + sys.platform + '.'
        print 'Python: {}, located at {}.'.format(sys.version[:5], sys.executable)
        sys.exit(1)

    # def useReact(self):
    #     return self._args.useReact

    def _install(self, package):
        try:
            print "Installing %s" % package
            subprocess.check_output(
                [sys.executable, "-m", "pip2", "-q", "install", "--user", package])
            return True
        except subprocess.CalledProcessError as ex:
            print "\033[91mWarn\033[00m: {} cannot be installed.\n Creating a non-react jemdoc website...".format(
                package)
            return False

    def _checkCssEngine(self, engine):
        engineName = engine.lower()
        if engineName not in ("less", "sass", "css", "scss"):
            raise argparse.ArgumentTypeError(
                "css engine '{}' is not supported".format(engine))
        return engineName

    def getCssEngine(self):
        return self._cssEngine

    def getCssStyle(self):
        return self._args.cssEngine.lower()

    @staticmethod
    def compileToCss(engine, inputfile, outputfile):
        if not os.path.exists(inputfile):
            print "\033[91mWarn\033[00m: %s does not exist in current directory, cannot be used!"
            return False

        try:
            out = os.path.join(os.getcwd(), outputfile)
            cmd = "{} {} {}".format(engine, inputfile, out)
            print cmd
            out = subprocess.check_output(cmd, shell=True)
            return True
        except subprocess.CalledProcessError as ex:
            print "\033[91mWarn\033[00m: {}".format(str(ex).lower())
            sys.exit(1)

    def _checkFileExists(self, filename):
        if "*" in filename:
            name, ext = os.path.splitext(filename)
            files = []
            if "*" in name:
                files = [str(f) for f in os.listdir('.') if os.path.isfile(f)]
            elif "*" in ext:
                files = [str(f) for f in os.listdir('.') if f.startswith(name)]
            return files

        if not os.path.exists(filename):
            raise argparse.ArgumentTypeError("file '{}' does not exists in current directory".format(filename))
        elif not os.path.isfile(filename):
            raise argparse.ArgumentTypeError("{} is a directory and not a file. Please provide a file".format(filename))
        return str(filename)

    def downloadStandardConfig(self):
        self._standardConfig.downloadStandardConfig()
        sys.exit(1)



if __name__ == "__main__":
    parser = CommandLineParser()
    args = parser.getArgs()
    if (args.info):
        parser.displayInfo()
    if (args.showConfig):
        parser.displayConfig()
    # print "Use React: {}".format(args.useReact)
    print "User Config file: {}".format(args.config)
    print "Outdir: {}".format(args.outdir)
    print "Innames: {}".format(args.input)
    print "CSS Engine: {}".format(args.cssEngine)




import pkgutil, os

class Style:
    defaultCss = """
/* Default css file for jemdoc. */

table#tlayout {
    border: none;
    border-collapse: separate;
    background: white;
}

body {
    background: white;
    font-family: Georgia, serif;
    padding-bottom: 8px;
    margin: 0;
}

#layout-menu {
    background: #f6f6f6;
    border: 1px solid #dddddd;
    padding-top: 0.5em;
    padding-left: 8px;
    padding-right: 8px;
    font-size: 1.0em;
    width: auto;
    white-space: nowrap;
    text-align: left;
    vertical-align: top;
}

#layout-menu td {
    background: #f4f4f4;
    vertical-align: top;
}

#layout-content {
    padding-top: 0.0em;
    padding-left: 1.0em;
    padding-right: 1.0em;
    border: none;
    background: white;
    text-align: left;
    vertical-align: top;
}

#layout-menu a {
    line-height: 1.5em;
    margin-left: 0.5em;
}

tt {
    background: #ffffdd;
}

pre, tt {
    font-size: 90%;
    font-family: monaco, monospace;
}

a, a > tt {
    color: #224b8d;
    text-decoration: none;
}

a:hover {
    border-bottom: 1px gray dotted;
}

#layout-menu a.current:link, #layout-menu a.current:visited {
    color: #022b6d;
    border-bottom: 1px gray solid;
}
#layout-menu a:link, #layout-menu a:visited, #layout-menu a:hover {
    color: #527bbd;
    text-decoration: none;
}
#layout-menu a:hover {
    text-decoration: none;
}

div.menu-category {
    border-bottom: 1px solid gray;
    margin-top: 0.8em;
    padding-top: 0.2em;
    padding-bottom: 0.1em;
    font-weight: bold;
}

div.menu-item {
    padding-left: 16px;
    text-indent: -16px;
}

div#toptitle {
    padding-bottom: 0.2em;
    margin-bottom: 1.5em;
    border-bottom: 3px double gray;
}

/* Reduce space if we begin the page with a title. */
div#toptitle + h2, div#toptitle + h3 {
    margin-top: -0.7em;
}

div#subtitle {
    margin-top: 0.0em;
    margin-bottom: 0.0em;
    padding-top: 0em;
    padding-bottom: 0.1em;
}

em {
    font-style: italic;
}

strong {
    font-weight: bold;
}


h1, h2, h3 {
    color: #527bbd;
    margin-top: 0.7em;
    margin-bottom: 0.3em;
    padding-bottom: 0.2em;
    line-height: 1.0;
    padding-top: 0.5em;
    border-bottom: 1px solid #aaaaaa;
}

h1 {
    font-size: 165%;
}

h2 {
    padding-top: 0.8em;
    font-size: 125%;
}

h2 + h3 {
    padding-top: 0.2em;
}

h3 {
    font-size: 110%;
    border-bottom: none;
}

p {
    margin-top: 0.0em;
    margin-bottom: 0.8em;
    padding: 0;
    line-height: 1.3;
}

pre {
    padding: 0;
    margin: 0;
}

div#footer {
    font-size: small;
    border-top: 1px solid #c0c0c0;
    padding-top: 0.1em;
    margin-top: 4.0em;
    color: #c0c0c0;
}

div#footer a {
    color: #80a0b0;
}

div#footer-text {
    float: left;
    padding-bottom: 8px;
}

ul, ol, dl {
    margin-top: 0.2em;
    padding-top: 0;
    margin-bottom: 0.8em;
}

dt {
    margin-top: 0.5em;
    margin-bottom: 0;
}

dl {
    margin-left: 20px;
}

dd {
    color: #222222;
}

dd > *:first-child {
    margin-top: 0;
}

ul {
    list-style-position: outside;
    list-style-type: square;
}

p + ul, p + ol {
    margin-top: -0.5em;
}

li ul, li ol {
    margin-top: -0.3em;
}

ol {
    list-style-position: outside;
    list-style-type: decimal;
}

li p, dd p {
    margin-bottom: 0.3em;
}


ol ol {
    list-style-type: lower-alpha;
}

ol ol ol {
    list-style-type: lower-roman;
}

p + div.codeblock {
    margin-top: -0.6em;
}

div.codeblock, div.infoblock {
    margin-right: 0%;
    margin-top: 1.2em;
    margin-bottom: 1.3em;
}

div.blocktitle {
    font-weight: bold;
    color: #cd7b62;
    margin-top: 1.2em;
    margin-bottom: 0.1em;
}

div.blockcontent {
    border: 1px solid silver;
    padding: 0.3em 0.5em;
}

div.infoblock > div.blockcontent {
    background: #ffffee;
}

div.blockcontent p + ul, div.blockcontent p + ol {
    margin-top: 0.4em;
}

div.infoblock p {
    margin-bottom: 0em;
}

div.infoblock li p, div.infoblock dd p {
    margin-bottom: 0.5em;
}

div.infoblock p + p {
    margin-top: 0.8em;
}

div.codeblock > div.blockcontent {
    background: #f6f6f6;
}

span.pycommand {
    color: #000070;
}

span.statement {
    color: #008800;
}
span.builtin {
    color: #000088;
}
span.special {
    color: #990000;
}
span.operator {
    color: #880000;
}
span.error {
    color: #aa0000;
}
span.comment, span.comment > *, span.string, span.string > * {
    color: #606060;
}

@media print {
    #layout-menu { display: none; }
}

#fwtitle {
    margin: 2px;
}

#fwtitle #toptitle {
    padding-left: 0.5em;
    margin-bottom: 0.5em;
}

#layout-content h1:first-child, #layout-content h2:first-child, #layout-content h3:first-child {
    margin-top: -0.7em;
}

div#toptitle h1, #layout-content div#toptitle h1 {
    margin-bottom: 0.0em;
    padding-bottom: 0.1em;
    padding-top: 0;
    margin-top: 0.5em;
    border-bottom: none;
}

table {
    border: 2px solid black;
    border-collapse: collapse;
}

td {
    padding: 2px;
    padding-left: 0.5em;
    padding-right: 0.5em;
    text-align: center;
    border: 1px solid gray;
}

table + table {
    margin-top: 1em;
}

tr.heading {
    font-weight: bold;
    border-bottom: 2px solid black;
}

img {
    border: none;
}

""" 

    @staticmethod
    def getDefaultCSS():
        return Style.defaultCss

    @staticmethod
    def downloadDefaultCSS(outdir):
        filePath = os.path.join(outdir, "jemdoc.css")
        file = open(filePath, "w")
        file.write(Style.defaultCss)
        file.flush()
        file.close()



class ControlStruct(object):
    def __init__(self, infile, outfile=None, conf=None, inname=None, outdirname='.'):
        self.inname = inname
        self.inf = infile
        self.outf = outfile
        self.conf = conf
        self.linenum = 0
        self.otherfiles = []
        self.texlines = []
        self.analytics = None
        self.baseline = None
        self.outdirname = outdirname

    def pushfile(self, newfile):
        """
    This function updates which file is self.inf
    Treats self.otherfiles as a stack with [0] being top
    """
        self.otherfiles.insert(0, self.inf)
        self.inf = open(newfile, 'rb')

    def nextfile(self):
        self.inf.close()
        self.inf = self.otherfiles.pop(0)

    def getCurrentInputFile(self):
        return self.inf

    def getOutputFile(self):
        return self.outf

    def getOutputDir(self):
        return self.outdirname


#/usr/bin/env python2

import setuptools

# long description
longDesc = ""
with open("README.md", "r") as f:
   longDesc = f.read()

# parsing requirements.txt
requirements = []
with open("requirements.txt", "r") as fr:
   for line in fr:
      requirements.append(line.strip())

# setup tools
setuptools.setup(name='jemdoc',
      version='0.8.0',
      description='jemdoc is a light text-based markup language designed for creating websites. It takes a text file written with jemdoc markup, an optional configuration file and an optional menu file, and makes static websites',
      author='David Sarpong',
      author_email='sarpong.david2@gmail.com',
      long_description=longDesc,
      long_description_content_type="text/markdown",
      python_requires="==2.7.*",
      url='https://github.com/sarpongdk/Jemdoc2.0',
      packages=setuptools.find_packages(),
      install_requires=requirements,
      classifiers=[
         "Programming Language :: Python :: 2",
         "Operating System :: OS Independent"
      ],
      entry_points={
         "console_scripts": ["jemdoc=jemdoc.main:main"]
      },
      include_package_data=True,
      package_data={
         # include any config/*.txt file and css/*.css file in jemdoc package
         "": [ "*.css" ]
      }
    )



import os

class DefaultConfig:
    def __init__(self):
        self._config = """
  [firstbit]
  <!doctype html>
  <html lang="en">
  <head>
  <meta charset="utf-8">
  <meta name="generator" content="jemdoc, see http://jemdoc.jaboc.net/" />
  <meta http-equiv="Content-Type" content="text/html;charset=utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
  <script type="text/x-mathjax-config">
  MathJax.Hub.Config({
    tex2jax: {
      inlineMath: [['$','$'], ['\\(','$\\)']],
      displayMath: [['$$','$$'], ['$\\[','$\\]']]
    },
    TeX: { 
      extensions: ["AMScd.js", "action.js", "autobold.js", "cancel.js", "begingroup.js", "color.js", "enclose.js"] 
    }
  });
  </script>
  <script type="text/javascript" async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/latest.js?config=TeX-MML-AM_CHTML"></script>
  <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/css/bootstrap.min.css" integrity="sha384-Gn5384xqQ1aoWXA+058RXPxPg6fy4IWvTNh0E263XmFcJlSAwiGgFAW/dAiS6JXm" crossorigin="anonymous">
  
  [defaultcss]
  <link rel="stylesheet" href="jemdoc.css" type="text/css" />
  
  [windowtitle]
  <title>|</title>

  [fwtitlestart]
  <div class="jumbotron text-center" id="fwtitle">

  [fwtitleend]
  </div>

  [navstart]
  <nav class="navbar navbar-expand-lg navbar-light bg-light %s" id="%s">
  <!-- <a class="navbar-brand" href="#">Navbar</a> = Navbar Brand -->
  <button class="navbar-toggler" type="button" data-toggle="collapse" data-target="#navbarNavAltMarkup" aria-controls="navbarNavAltMarkup" aria-expanded="false" aria-label="Toggle navigation">
  <span class="navbar-toggler-icon"></span>
  </button>
  <div class="collapse navbar-collapse" id="navbarNavAltMarkup">
  <div class="navbar-nav nav-fill w-100">

  [navend]
  </div>
  </div>
  </nav>

  [currentnavitem]
  <a class="nav-item nav-link active" href="|1">|2</a>

  [navitem]
  <a class="nav-item nav-link" href="|1">|2</a>

  [nonav]
  <div class="title-divider"></div>

  [doctitle]
  <div class="toptitle">
  <h1>|</h1>
  
  [subtitle]
  <h2 id="subtitle">|</h2>
  
  [doctitleend]
  </div>
  
  [bodystart]
  </head>
  <body>
  
  [analytics]
  <script type="text/javascript">
  var gaJsHost = (("https:" == document.location.protocol) ? "https://ssl." : "http://www.");
  document.write(unescape("%3Cscript src='" + gaJsHost + "google-analytics.com/ga.js' type='text/javascript'%3E%3C/script%3E"));
  </script>
  <script type="text/javascript">
  try {
      var pageTracker = _gat._getTracker("|");
      pageTracker._trackPageview();
  } catch(err) {}</script>
  
  [menustart]
  <main class="container" id="tlayout">
  <div class="row">
  <!-- Side menu -->
  <aside class="col-12 col-md-3 %s" id="%s"> 
  <ul class="nav nav-pills flex-column">
  
  [menuend]
  </ul>
  </aside>
  <div class="col-12 col-md-9" id="main-content">
  
  [menucategory]
  <div class="menu-category">|</div>

  [menuitem]
  <li class="nav-item menu-item"><a href="|1" class="nav-link">|2</a></li>

  [specificcss]
  <link rel="stylesheet" href="|" type="text/css" />

  [specificjs]
  <script src="|.js" type="text/javascript"></script>
  
  [currentmenuitem]
  <li class="nav-item menu-item"><a href="|1" class="nav-link active">|2</a></li>
  
  [nomenu]
  <div id="main-content">
  
  [menulastbit]
  <!-- End of main body -->
  </div>
  </div>
  </main>
  
  [nomenulastbit]
  </div>
  
  [bodyend]
  <script src="https://code.jquery.com/jquery-3.2.1.slim.min.js" integrity="sha384-KJ3o2DKtIkvYIK3UENzmM7KCkRr/rE9/Qpg6aAZGJwFDMVNA/GpGFF93hXpG5KkN" crossorigin="anonymous"></script>
  <script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.12.9/umd/popper.min.js" integrity="sha384-ApNbgh9B+Y1QKtv3Rn7W3mgPxhU9K/ScQsAP7hUibX39j7fakFPskvXusvfa0b4Q" crossorigin="anonymous"></script>
  <script src="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/js/bootstrap.min.js" integrity="sha384-JZR6Spejh4U02d8jOt6vLEHfe/JQGiRRSQQxSfFWpi1MquVdAyjUar5+76PVCmYl" crossorigin="anonymous"></script>
  </body>
  </html>
  
  [infoblock]
  <div class="infoblock">
  
  [codeblock]
  <div class="codeblock">
  
  [blocktitle]
  <div class="blocktitle">|</div>
  
  [infoblockcontent]
  <div class="blockcontent">
  
  [codeblockcontent]
  <div class="blockcontent"><pre>
  
  [codeblockend]
  </pre></div></div>
  
  [codeblockcontenttt]
  <div class="blockcontent"><tt class="tthl">
  
  [codeblockendtt]
  </tt></div></div>
  
  [infoblockend]
  </div></div>
  
  [footerstart]
  <footer class="jumbotron" id="footer">
  <div id="footer footer-text">
  
  [footerend]
  </div>
  </footer>
  
  [lastupdated]
  Page generated |, by <a href="http://jemdoc.jaboc.net/">jemdoc</a>.

  [sourcelink]
  (<a href="|">source</a>)

  """

    def downloadStandardConfig(self):
        filename = "standardconfig.txt"
        filePath = os.path.join(".", filename)
        fileObj = open(filePath, "w")
        print "Downloading standard configuration file %s..." % filename
        fileObj.write(self._config)
        fileObj.flush()
        fileObj.close()

    def getStandardConfig(self):
        return self._config